@@ -5,7 +5,6 @@ extends 'YAML::Loader::Base';
 
 use YAML::Loader::Base;
 use YAML::Types;
-use YAML::Node;
 
 # Context constants
 use constant LEAF       => 1;
@@ -35,7 +34,9 @@ sub _parse {
     $self->line(0);
     $self->die('YAML_PARSE_ERR_BAD_CHARS')
       if $self->stream =~ /$ESCAPE_CHAR/;
-    $self->{stream} =~ s/(.)\n\Z/$1/s;
+    $self->die('YAML_PARSE_ERR_NO_FINAL_NEWLINE')
+      if length($self->stream) and
+         $self->{stream} !~ s/(.)\n\Z/$1/s;
     $self->lines([split /\x0a/, $self->stream, -1]);
     $self->line(1);
     # Throw away any comments or blanks before the header (or start of
@@ -43,7 +44,6 @@ sub _parse {
     $self->_parse_throwaway_comments();
     $self->document(0);
     $self->documents([]);
-    $self->zero_indent([]);
     # Add an "assumed" header if there is no header and the stream is
     # not empty (after initial throwaways).
     if (not $self->eos) {
@@ -62,25 +62,17 @@ sub _parse {
         $self->offset->[0] = -1;
 
         if ($self->lines->[0] =~ /^---\s*(.*)$/) {
-            my @words = split /\s/, $1;
+            my @words = split /\s+/, $1;
             %directives = ();
-            while (@words) {
-                if ($words[0] =~ /^#(\w+):(\S.*)$/) {
-                    my ($key, $value) = ($1, $2);
-                    shift(@words);
-                    if (defined $directives{$key}) {
-                        $self->warn('YAML_PARSE_WARN_MULTIPLE_DIRECTIVES',
-                          $key, $self->document);
-                        next;
-                    }
-                    $directives{$key} = $value;
-                }
-                elsif ($words[0] eq '') {
-                    shift @words;
-                }
-                else {
-                    last;
+            while (@words && $words[0] =~ /^#(\w+):(\S.*)$/) {
+                my ($key, $value) = ($1, $2);
+                shift(@words);
+                if (defined $directives{$key}) {
+                    $self->warn('YAML_PARSE_WARN_MULTIPLE_DIRECTIVES',
+                      $key, $self->document);
+                    next;
                 }
+                $directives{$key} = $value;
             }
             $self->preface(join ' ', @words);
         }
@@ -119,8 +111,8 @@ sub _parse_node {
     my $self = shift;
     my $preface = $self->preface;
     $self->preface('');
-    my ($node, $type, $indicator, $chomp, $parsed_inline) = ('') x 5;
-    my ($anchor, $alias, $explicit, $implicit) = ('') x 4;
+    my ($node, $type, $indicator, $escape, $chomp) = ('') x 5;
+    my ($anchor, $alias, $explicit, $implicit, $class) = ('') x 5;
     ($anchor, $alias, $explicit, $implicit, $preface) =
       $self->_parse_qualifiers($preface);
     if ($anchor) {
@@ -128,21 +120,13 @@ sub _parse_node {
     }
     $self->inline('');
     while (length $preface) {
-        if ($preface =~ s/^($FOLD_CHAR|$LIT_CHAR_RX)//) {
+        my $line = $self->line - 1;
+        if ($preface =~ s/^($FOLD_CHAR|$LIT_CHAR_RX)(-|\+)?\d*\s*//) {
             $indicator = $1;
-            if ($preface =~ s/^([+-])[0-9]*//) {
-                $chomp = $1;
-            }
-            elsif ($preface =~ s/^[0-9]+([+-]?)//) {
-                $chomp = $1;
-            }
-            if ($preface =~ s/^(?:\s+#.*$|\s*)$//) {
-            }
-            else {
-                $self->die('YAML_PARSE_ERR_TEXT_AFTER_INDICATOR');
-            }
+            $chomp = $2 if defined($2);
         }
         else {
+            $self->die('YAML_PARSE_ERR_TEXT_AFTER_INDICATOR') if $indicator;
             $self->inline($preface);
             $preface = '';
         }
@@ -160,7 +144,6 @@ sub _parse_node {
     }
     elsif (length $self->inline) {
         $node = $self->_parse_inline(1, $implicit, $explicit);
-        $parsed_inline = 1;
         if (length $self->inline) {
             $self->die('YAML_PARSE_ERR_SINGLE_LINE');
         }
@@ -202,7 +185,17 @@ sub _parse_node {
     $#{$self->offset} = $self->level;
 
     if ($explicit) {
-        $node = $self->_parse_explicit($node, $explicit) if !$parsed_inline;
+        if ($class) {
+            if (not ref $node) {
+                my $copy = $node;
+                undef $node;
+                $node = \$copy;
+            }
+            CORE::bless $node, $class;
+        }
+        else {
+            $node = $self->_parse_explicit($node, $explicit);
+        }
     }
     if ($anchor) {
         if (ref($self->anchor2node->{$anchor}) eq 'YAML-anchor2node') {
@@ -225,6 +218,7 @@ sub _parse_qualifiers {
     my ($anchor, $alias, $explicit, $implicit, $token) = ('') x 5;
     $self->inline('');
     while ($preface =~ /^[&*!]/) {
+        my $line = $self->line - 1;
         if ($preface =~ s/^\!(\S+)\s*//) {
             $self->die('YAML_PARSE_ERR_MANY_EXPLICIT') if $explicit;
             $explicit = $1;
@@ -233,15 +227,15 @@ sub _parse_qualifiers {
             $self->die('YAML_PARSE_ERR_MANY_IMPLICIT') if $implicit;
             $implicit = 1;
         }
-        elsif ($preface =~ s/^\&([^ ,:]*)\s*//) {
+        elsif ($preface =~ s/^\&([^ ,:]+)\s*//) {
             $token = $1;
             $self->die('YAML_PARSE_ERR_BAD_ANCHOR')
-              unless $token =~ /^[a-zA-Z0-9_.\/-]+$/;
+              unless $token =~ /^[a-zA-Z0-9]+$/;
             $self->die('YAML_PARSE_ERR_MANY_ANCHOR') if $anchor;
             $self->die('YAML_PARSE_ERR_ANCHOR_ALIAS') if $alias;
             $anchor = $token;
         }
-        elsif ($preface =~ s/^\*([^ ,:]*)\s*//) {
+        elsif ($preface =~ s/^\*([^ ,:]+)\s*//) {
             $token = $1;
             $self->die('YAML_PARSE_ERR_BAD_ALIAS')
               unless $token =~ /^[a-zA-Z0-9]+$/;
@@ -276,7 +270,7 @@ sub _parse_explicit {
             $node = \$value;
         }
 
-        if ( length($class) and $YAML::LoadBlessed ) {
+        if ( length($class) ) {
             CORE::bless($node, $class);
         }
 
@@ -302,7 +296,7 @@ sub _parse_explicit {
             require YAML::Node;
             return $class->yaml_load(YAML::Node->new($node, $explicit));
         }
-        elsif ($YAML::LoadBlessed) {
+        else {
             if (ref $node) {
                 return CORE::bless $node, $class;
             }
@@ -310,9 +304,6 @@ sub _parse_explicit {
                 return CORE::bless \$node, $class;
             }
         }
-        else {
-            return $node;
-        }
     }
     elsif (ref $node) {
         require YAML::Node;
@@ -329,7 +320,7 @@ sub _parse_explicit {
 sub _parse_mapping {
     my $self = shift;
     my ($anchor) = @_;
-    my $mapping = $self->preserve ? YAML::Node->new({}) : {};
+    my $mapping = {};
     $self->anchor2node->{$anchor} = $mapping;
     my $key;
     while (not $self->done and $self->indent == $self->offset->[$self->level]) {
@@ -341,11 +332,11 @@ sub _parse_mapping {
             $key = "$key";
         }
         # If "default" key (equals sign)
-        elsif ($self->{content} =~ s/^\=\s*(?=:)//) {
+        elsif ($self->{content} =~ s/^\=\s*//) {
             $key = VALUE;
         }
         # If "comment" key (slash slash)
-        elsif ($self->{content} =~ s/^\=\s*(?=:)//) {
+        elsif ($self->{content} =~ s/^\=\s*//) {
             $key = COMMENT;
         }
         # Regular scalar key:
@@ -357,21 +348,15 @@ sub _parse_mapping {
             $self->inline('');
         }
 
-        unless ($self->{content} =~ s/^:(?:\s+#.*$|\s*)//) {
+        unless ($self->{content} =~ s/^:\s*//) {
             $self->die('YAML_LOAD_ERR_BAD_MAP_ELEMENT');
         }
         $self->preface($self->content);
-        my $level = $self->level;
-
-        # we can get a zero indented sequence, possibly
-        my $zero_indent = $self->zero_indent;
-        $zero_indent->[ $level ] = 0;
+        my $line = $self->line;
         $self->_parse_next_line(COLLECTION);
         my $value = $self->_parse_node();
-        $#$zero_indent = $level;
-
         if (exists $mapping->{$key}) {
-            $self->warn('YAML_LOAD_WARN_DUPLICATE_KEY', $key);
+            $self->warn('YAML_LOAD_WARN_DUPLICATE_KEY');
         }
         else {
             $mapping->{$key} = $value;
@@ -391,33 +376,9 @@ sub _parse_seq {
             $self->preface(defined($1) ? $1 : '');
         }
         else {
-            if ($self->zero_indent->[ $self->level ]) {
-                last;
-            }
             $self->die('YAML_LOAD_ERR_BAD_SEQ_ELEMENT');
         }
-
-        # Check whether the preface looks like a YAML mapping ("key: value").
-        # This is complicated because it has to account for the possibility
-        # that a key is a quoted string, which itself may contain escaped
-        # quotes.
-        my $preface = $self->preface;
-        if ($preface =~ m/^ (\s*) ( - (?: \ .* | $ ) ) /x) {
-            $self->indent($self->offset->[$self->level] + 2 + length($1));
-            $self->content($2);
-            $self->level($self->level + 1);
-            $self->offset->[$self->level] = $self->indent;
-            $self->preface('');
-            push @$seq, $self->_parse_seq('');
-            $self->{level}--;
-            $#{$self->offset} = $self->level;
-        }
-        elsif (
-             $preface =~ /^ (\s*) ((') (?:''|[^'])*? ' \s* \: (?:\ |$).*) $/x or
-             $preface =~ /^ (\s*) ((") (?:\\\\|[^"])*? " \s* \: (?:\ |$).*) $/x or
-             $preface =~ /^ (\s*) (\?.*$)/x or
-             $preface =~ /^ (\s*) ([^\s:#&!\[\]\{\},*|>].*\:(\ .*|$))/x
-           ) {
+        if ($self->preface =~ /^(\s*)(\w.*\:(?: |$).*)$/) {
             $self->indent($self->offset->[$self->level] + 2 + length($1));
             $self->content($2);
             $self->level($self->level + 1);
@@ -485,11 +446,6 @@ sub _parse_inline {
             $node = $self->_parse_inline_simple();
         }
         $node = $self->_parse_implicit($node) unless $explicit;
-
-        if ($self->numify and defined $node and not ref $node and length $node
-            and $node =~ m/\A-?(?:0|[1-9][0-9]*)?(?:\.[0-9]*)?(?:[eE][+-]?[0-9]+)?\z/) {
-            $node += 0;
-        }
     }
     if ($explicit) {
         $node = $self->_parse_explicit($node, $explicit);
@@ -516,13 +472,13 @@ sub _parse_inline_mapping {
 
     $self->die('YAML_PARSE_ERR_INLINE_MAP')
       unless $self->{inline} =~ s/^\{\s*//;
-    while (not $self->{inline} =~ s/^\s*\}(\s+#.*$|\s*)//) {
+    while (not $self->{inline} =~ s/^\s*\}//) {
         my $key = $self->_parse_inline();
         $self->die('YAML_PARSE_ERR_INLINE_MAP')
           unless $self->{inline} =~ s/^\: \s*//;
         my $value = $self->_parse_inline();
         if (exists $node->{$key}) {
-            $self->warn('YAML_LOAD_WARN_DUPLICATE_KEY', $key);
+            $self->warn('YAML_LOAD_WARN_DUPLICATE_KEY');
         }
         else {
             $node->{$key} = $value;
@@ -543,7 +499,7 @@ sub _parse_inline_seq {
 
     $self->die('YAML_PARSE_ERR_INLINE_SEQUENCE')
       unless $self->{inline} =~ s/^\[\s*//;
-    while (not $self->{inline} =~ s/^\s*\](\s+#.*$|\s*)//) {
+    while (not $self->{inline} =~ s/^\s*\]//) {
         my $value = $self->_parse_inline();
         push @$node, $value;
         next if $self->inline =~ /^\s*\]/;
@@ -556,43 +512,33 @@ sub _parse_inline_seq {
 # Parse the inline double quoted string.
 sub _parse_inline_double_quoted {
     my $self = shift;
-    my $inline = $self->inline;
-    if ($inline =~ s/^"//) {
-        my $node = '';
-
-        while ($inline =~ s/^(\\.|[^"\\]+)//) {
-            my $capture = $1;
-            $capture =~ s/^\\"/"/;
-            $node .= $capture;
-            last unless length $inline;
-        }
-        if ($inline =~ s/^"(?:\s+#.*|\s*)//) {
-            $self->inline($inline);
-            return $node;
-        }
+    my $node;
+    # https://rt.cpan.org/Public/Bug/Display.html?id=90593
+    if ($self->inline =~ /^"((?:(?:\\"|[^"]){0,32766}){0,32766})"\s*(.*)$/) {
+        $node = $1;
+        $self->inline($2);
+        $node =~ s/\\"/"/g;
     }
-    $self->die('YAML_PARSE_ERR_BAD_DOUBLE');
+    else {
+        $self->die('YAML_PARSE_ERR_BAD_DOUBLE');
+    }
+    return $node;
 }
 
 
 # Parse the inline single quoted string.
 sub _parse_inline_single_quoted {
     my $self = shift;
-    my $inline = $self->inline;
-    if ($inline =~ s/^'//) {
-        my $node = '';
-        while ($inline =~ s/^(''|[^']+)//) {
-            my $capture = $1;
-            $capture =~ s/^''/'/;
-            $node .= $capture;
-            last unless length $inline;
-        }
-        if ($inline =~ s/^'(?:\s+#.*|\s*)//) {
-            $self->inline($inline);
-            return $node;
-        }
+    my $node;
+    if ($self->inline =~ /^'((?:(?:''|[^']){0,32766}){0,32766})'\s*(.*)$/) {
+        $node = $1;
+        $self->inline($2);
+        $node =~ s/''/'/g;
     }
-    $self->die('YAML_PARSE_ERR_BAD_SINGLE');
+    else {
+        $self->die('YAML_PARSE_ERR_BAD_SINGLE');
+    }
+    return $node;
 }
 
 # Parse the inline unquoted string and do implicit typing.
@@ -612,14 +558,11 @@ sub _parse_inline_simple {
 sub _parse_implicit {
     my $self = shift;
     my ($value) = @_;
-    # remove trailing comments and whitespace
-    $value =~ s/^#.*$//;
-    $value =~ s/\s+#.*$//;
     $value =~ s/\s*$//;
     return $value if $value eq '';
     return undef if $value =~ /^~$/;
     return $value
-      unless $value =~ /^[\@\`]/ or
+      unless $value =~ /^[\@\`\^]/ or
              $value =~ /^[\-\?]\s/;
     $self->die('YAML_PARSE_ERR_BAD_IMPLICIT', $value);
 }
@@ -687,28 +630,23 @@ sub _parse_next_line {
     $self->die('YAML_EMIT_ERR_BAD_LEVEL') unless defined $offset;
     shift @{$self->lines};
     $self->eos($self->{done} = not @{$self->lines});
-    if ($self->eos) {
-        $self->offset->[$level + 1] = $offset + 1;
-        return;
-    }
+    return if $self->eos;
     $self->{line}++;
 
     # Determine the offset for a new leaf node
-    # TODO
     if ($self->preface =~
-        qr/(?:^|\s)(?:$FOLD_CHAR|$LIT_CHAR_RX)(?:[+-]([0-9]*)|([0-9]*)[+-]?)(?:\s+#.*|\s*)$/
+        qr/(?:^|\s)(?:$FOLD_CHAR|$LIT_CHAR_RX)(?:-|\+)?(\d*)\s*$/
        ) {
-        my $explicit_indent = length $1 ? $1 : length $2 ? $2 : '';
         $self->die('YAML_PARSE_ERR_ZERO_INDENT')
-          if length($explicit_indent) and $explicit_indent == 0;
+          if length($1) and $1 == 0;
         $type = LEAF;
-        if (length($explicit_indent)) {
-            $self->offset->[$level + 1] = $offset + $explicit_indent;
+        if (length($1)) {
+            $self->offset->[$level + 1] = $offset + $1;
         }
         else {
             # First get rid of any comments.
             while (@{$self->lines} && ($self->lines->[0] =~ /^\s*#/)) {
-                $self->lines->[0] =~ /^( *)/;
+                $self->lines->[0] =~ /^( *)/ or die;
                 last unless length($1) <= $offset;
                 shift @{$self->lines};
                 $self->{line}++;
@@ -728,22 +666,10 @@ sub _parse_next_line {
     elsif ($type == COLLECTION and
            $self->preface =~ /^(\s*(\!\S*|\&\S+))*\s*$/) {
         $self->_parse_throwaway_comments();
-        my $zero_indent = $self->zero_indent;
         if ($self->eos) {
             $self->offset->[$level+1] = $offset + 1;
             return;
         }
-        elsif (
-            defined $zero_indent->[ $level ]
-            and not $zero_indent->[ $level ]
-            and $self->lines->[0] =~ /^( {$offset,})-(?: |$)/
-        ) {
-            my $new_offset = length($1);
-            $self->offset->[$level+1] = $new_offset;
-            if ($new_offset == $offset) {
-                $zero_indent->[ $level+1 ] = 1;
-            }
-        }
         else {
             $self->lines->[0] =~ /^( *)\S/ or
                 $self->die('YAML_PARSE_ERR_NONSPACE_INDENTATION');
@@ -758,22 +684,12 @@ sub _parse_next_line {
     }
 
     if ($type == LEAF) {
-        if (@{$self->lines} and
+        while (@{$self->lines} and
                $self->lines->[0] =~ m{^( *)(\#)} and
                length($1) < $offset
               ) {
-            if ( length($1) < $offset) {
-                shift @{$self->lines};
-                $self->{line}++;
-                # every comment after that is also thrown away regardless
-                # of identation
-                while (@{$self->lines} and
-                       $self->lines->[0] =~ m{^( *)(\#)}
-                      ) {
-                    shift @{$self->lines};
-                    $self->{line}++;
-                }
-            }
+            shift @{$self->lines};
+            $self->{line}++;
         }
         $self->eos($self->{done} = not @{$self->lines});
     }
